/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2018                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                              HC_KEY1                                   **
**                      Wersja z dnia : 2018.07.09                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include "key1eeprom.h"

#define WatchDogActive

#ifdef WatchDogActive
#define WatchdogReset                  wdt_reset ( )
#define WatchdogOn                     wdt_enable ( WDTO_1S )
#define WatchdogOff                    wdt_disable ( )
#else
#define WatchdogReset
#define WatchdogOn
#define WatchdogOff
#endif


#define false                          0
#define true                           1

#define StatusLed                      0
#define RxTxLed                        2
#define KeyLed                         3    // 1

#define KeyPin                         5              // PORTD 5
#define KeyPort                        PORTD
#define KeyPortDirection               DDRD
#define KeyInputReg                    PIND

#define TimerCounterDelayLimit         3000  // 2048
#define StandardLEDPulseDelay          832
#define TimerPrescaller                3600

#define LEDPort                        PORTC
#define LEDPortDirection               DDRC

#define RxDPin                         0              // PORTD . 0
#define TxDPin                         1              // PORTD . 1
#define RxTxDirPin                     2              // PORTD . 2

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define StartActiveKeyCode             0x11
#define StopActiveKeyCode              0x22

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXDBuffSize                    128
#define RXDBuffSize                    32

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t State ;
                 uint8_t Buffer [ TXDBuffSize ] ;
               } TXDRecT ;
typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ RXDBuffSize ] ;
               } RXDRecT ;

#define CyclBuffSize                   256

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ CyclBuffSize ] ;
               } CyclicRecT ;

typedef void ( * OperationProcT ) ( void ) ;

typedef struct {
                 volatile TXDRecT          TXDRec ;
                 volatile RXDRecT          RXDRec ;
                 OperationProcT            SerialService ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
                 uint16_t                  SerialLEDDelay ;
               } ServiceType ;

typedef ServiceType * PtrServiceType ;

#define TempEchoDevicesArraySize       4

typedef enum { KeybIdle ,
               KeybWaitOnPress ,
               KeybReady ,
               KeybActive ,
               KeybWaitOnFree ,
               KeybWait
             } KeybStatusType ;

#define KeybDelayLimit                 2

static ServiceType SerialService ;
static KeybStatusType KeyStatus ;
static uint16_t KeyDelayCt ;
static uint16_t KeybDelayTimer ;
static uint8_t KeybCodeBuffer ;
static uint8_t ConnectContext ;
static volatile uint8_t TimeEvent ;
static uint16_t TimerCounter ;
static DeviceIDType TempEchoDevices [ TempEchoDevicesArraySize ] ;
static CyclicRecT CyclicBuffer ;


static const uint16_t CRCTable [ 16 ] = { 0x0000 ,  /* 00 */
                                          0xCC01 ,  /* 01 */
                                          0xD801 ,  /* 02 */
                                          0x1400 ,  /* 03 */
                                          0xF001 ,  /* 04 */
                                          0x3C00 ,  /* 05 */
                                          0x2800 ,  /* 06 */
                                          0xE401 ,  /* 07 */
                                          0xA001 ,  /* 08 */
                                          0x6C00 ,  /* 09 */
                                          0x7800 ,  /* 10 */
                                          0xB401 ,  /* 11 */
                                          0x5000 ,  /* 12 */
                                          0x9C01 ,  /* 13 */
                                          0x8801 ,  /* 14 */
                                          0x4400 } ;/* 15 */

static uint8_t HardwareSoftwareVersion [ ] PROGMEM   = "HVER*1_00*V-2013.SVER*1_1*VII-2018\r" ;
static uint8_t DeviceName [ ] PROGMEM                = "HCKEY1-czytnik stanow\r" ;
static uint8_t MyDefaultID [ ] PROGMEM               = "KEY1" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM     = "GENERAL" ;
static uint8_t OK [ ] PROGMEM                        = "<OK>\r" ;
static uint8_t Fail [ ] PROGMEM                      = "<FA>\r" ;
static uint8_t HexalTable [ ] PROGMEM                = "0123456789ABCDEF" ;
static uint8_t NoDataText [ ] PROGMEM                = "NODATA\r" ;
static uint8_t EmptyFlagTxt [ ] PROGMEM              = "<EMPTY>\r" ;
static uint8_t StartSelector [ ] PROGMEM             = "START" ;
static uint8_t StopSelector [ ] PROGMEM              = "STOP" ;
static uint8_t ChannelIsOnState [ ] PROGMEM          = "CHISON" ;
static uint8_t ChannelIsOffState [ ] PROGMEM         = "CHISOF" ;

static uint8_t ReplyCommand [ ] PROGMEM              = "REPLY" ;
static uint8_t ConnectCommand [ ] PROGMEM            = "CONNECT" ;
static uint8_t DisConnectCommand [ ] PROGMEM         = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM      = "SETMID" ;
static uint8_t HelloCommand [ ] PROGMEM              = "HELLO" ;
static uint8_t GetVersionCommand [ ] PROGMEM         = "GETVER" ;
static uint8_t ReinitCommand [ ] PROGMEM             = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM            = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM             = "SAVEEE" ;
static uint8_t AddSuperiorChnCommand [ ] PROGMEM     = "ADDSUP" ;
static uint8_t DelSuperiorChnCommand [ ] PROGMEM     = "DELSUP" ;
static uint8_t ListSuperiorChnCommand [ ] PROGMEM    = "LISTSUP" ;
static uint8_t ChStateServiceOnCommand [ ] PROGMEM   = "CSSERVON" ;
static uint8_t ChStateServiceOffCommand [ ] PROGMEM  = "CSSERVOFF" ;
static uint8_t ChStateServiceListCommand [ ] PROGMEM = "CSSERVLIST" ;
static uint8_t NewPasswordCommand [ ] PROGMEM        = "NEWPW" ;
static uint8_t AddKeyCommand [ ] PROGMEM             = "ADDKC" ;
static uint8_t DelKeyCommand [ ] PROGMEM             = "DELKC" ;
static uint8_t ListKeyCommand [ ] PROGMEM            = "LISTKC" ;
static uint8_t AddStatusChnCommand [ ] PROGMEM       = "ADDSTA" ;
static uint8_t DelStatusChnCommand [ ] PROGMEM       = "DELSTA" ;
static uint8_t ListStatusChnCommand [ ] PROGMEM      = "LISTSTA" ;

typedef void ( * CommandExecuteServiceProcT ) ( void ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void ReplyCommandExecute ( void ) ;
static void ConnectCommandExecute ( void ) ;
static void DisConnectCommandExecute ( void ) ;
static void SetDeviceNameCommandExecute ( void ) ;
static void HelloCommandExecute ( void ) ;
static void GetVersionCommandExecute ( void ) ;
static void ReinitCommandExecute ( void ) ;
static void ResetEECommandExecute ( void ) ;
static void SaveEECommandExecute ( void ) ;
static void AddSuperiorChnCommandExecute ( void ) ;
static void DelSuperiorChnCommandExecute ( void ) ;
static void ChStateServiceOnCommandExecute ( void ) ;
static void ChStateServiceOffCommandExecute ( void ) ;
static void ChStateServiceListCommandExecute ( void ) ;
static void ListSuperiorChnCommandExecute ( void ) ;
static void NewPasswordCommandExecute ( void ) ;
static void AddKeyCommandExecute ( void ) ;
static void DelKeyCommandExecute ( void ) ;
static void ListKeyCommandExecute ( void ) ;
static void AddStatusChnCommandExecute ( void ) ;
static void DelStatusChnCommandExecute ( void ) ;
static void ListStatusChnCommandExecute ( void ) ;

#define CommandNameAssociationTabeSize               22

static CommandNameAssociationRecT CommandNameAssociationTabe [ CommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ReplyCommandExecute                 , ReplyCommand } ,
/* 01 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 02 */ { DisConnectCommandExecute            , DisConnectCommand } ,
/* 03 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 04 */ { HelloCommandExecute                 , HelloCommand } ,
/* 05 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 06 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 07 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 08 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 09 */ { AddSuperiorChnCommandExecute        , AddSuperiorChnCommand } ,
/* 10 */ { DelSuperiorChnCommandExecute        , DelSuperiorChnCommand } ,
/* 11 */ { ListSuperiorChnCommandExecute       , ListSuperiorChnCommand } ,
/* 12 */ { ChStateServiceOnCommandExecute      , ChStateServiceOnCommand } ,
/* 13 */ { ChStateServiceOffCommandExecute     , ChStateServiceOffCommand } ,
/* 14 */ { ChStateServiceListCommandExecute    , ChStateServiceListCommand } ,
/* 15 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 16 */ { AddKeyCommandExecute                , AddKeyCommand } ,
/* 17 */ { DelKeyCommandExecute                , DelKeyCommand } ,
/* 18 */ { ListKeyCommandExecute               , ListKeyCommand } ,
/* 19 */ { AddStatusChnCommandExecute          , AddStatusChnCommand } ,
/* 20 */ { DelStatusChnCommandExecute          , DelStatusChnCommand } ,
/* 21 */ { ListStatusChnCommandExecute         , ListStatusChnCommand } ,
} ;


static uint8_t KeyPressed ( void )
{
  /*----------------------------------------------------------------*/
  if ( KeybCodeBuffer )
    return true ;
  else
    return false ;
} /* KeyPressed */


static uint8_t ReadKey ( void )
{
  uint8_t Tmp ;
  /*----------------------------------------------------------------*/
  Tmp = KeybCodeBuffer ;
  KeybCodeBuffer = 0 ;
  return Tmp ;
} /* ReadKey */


static uint8_t HardwareKeybRead ( void )
{
  uint8_t KeybData ;
  /*----------------------------------------------------------------*/
  KeybData = KeyInputReg ^ 0xFF ;
  return ( KeybData & ( 1 << KeyPin ) ) ;
} /* HardwareKeybRead */


static void KeyService ( void )
{
  uint8_t KeybActiveStatus ;
  /*----------------------------------------------------------------*/
  KeybActiveStatus = HardwareKeybRead ( ) ;
  if ( KeybActiveStatus )
    LEDPort &= ~ ( 1 << KeyLed ) ;
  else
    LEDPort |= 1 << KeyLed ;
  switch ( KeyStatus )
  {
    case KeybIdle        :
      if ( KeybActiveStatus )
      {
        KeyStatus = KeybWaitOnPress ;
        KeyDelayCt = KeybDelayLimit ;
      } /* if */ ;
      break ;
    case KeybWaitOnPress :
      KeyDelayCt -- ;
      if ( ! KeyDelayCt )
      {
        if ( KeybActiveStatus )
          KeyStatus = KeybReady ;
        else
          KeyStatus = KeybIdle ;
      } /* if */ ;
      break ;
    case KeybReady       :
      KeyStatus = KeybActive ;
      break ;
    case KeybActive      :
      KeybCodeBuffer = StartActiveKeyCode ;
      KeyStatus = KeybWaitOnFree ;
      break ;
    case KeybWaitOnFree  :
      if ( ! KeybActiveStatus )
      {
        KeybCodeBuffer = StopActiveKeyCode ;
        KeyStatus = KeybWait ;
      } /* if */ ;
      break ;
    case KeybWait        :
      if ( ! KeybActiveStatus )
      {
        KeyStatus = KeybIdle ;
      } /* if */ ;
      break ;
  } /* switch */ ;
} /* KeyService */


static void USARTReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  PORTD &= ~ ( 1 << RxTxDirPin ) ;
} /* USARTReceiveMode */


static void USARTTransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  PORTD |= ( 1 << RxTxDirPin ) ;
} /* USARTTransmitMode */


SIGNAL ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    TimeEvent = true ;
  } /* if */ ;
  if ( SerialService . SerialLEDDelay )
  {
    SerialService . SerialLEDDelay -- ;
    if ( ! SerialService . SerialLEDDelay )
    {
      LEDPort |= ( 1 << RxTxLed ) ;
    } /* if */ ;
  } /* if */ ;
  KeybDelayTimer ++ ;
  if ( KeybDelayTimer > 100 )
  {
    KeybDelayTimer = 0 ;
    KeyService ( ) ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void StartRxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDPort &= ~ ( 1 << RxTxLed ) ;
  SerialService . SerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRxTxLED */


SIGNAL ( USART_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR0 ;
  SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycTo ] = RecCh ;
  SerialService . RXDRec . CycTo ++ ;
  if ( SerialService . RXDRec . CycTo >= RXDBuffSize )
    SerialService . RXDRec . CycTo = 0 ;
  StartRxTxLED ( ) ;
} /* USART_RX_vect */


SIGNAL ( USART_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . TXDRec . CycFr == SerialService . TXDRec . CycTo )
  {
    SerialService . TXDRec . State = IdleState ;
    USARTReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR0 = SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycFr ] ;
    SerialService . TXDRec . CycFr ++ ;
    if ( SerialService . TXDRec . CycFr >= TXDBuffSize )
      SerialService . TXDRec . CycFr = 0 ;
  } /* if ... else */ ;
} /* USART_TX_vect */


static void SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  USARTTransmitMode ( ) ;
  StartRxTxLED ( ) ;
  cli ( ) ;
  if ( SerialService . TXDRec . State == IdleState )
  {
    SerialService . TXDRec . State = BusyState ;
    UDR0 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = SerialService . TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != SerialService . TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      WatchdogReset ;
      cli ( ) ;
    } /* for */ ;
    SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycTo ] = SerialCh ;
    SerialService . TXDRec . CycTo ++ ;
    if ( SerialService . TXDRec . CycTo >= TXDBuffSize )
      SerialService . TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* SendSerial */


static void SendSerialTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* SendSerialTxt */


static void SendSerialFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* SendSerialFlashTxt */


static uint8_t SerialDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . RXDRec . CycFr == SerialService . RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* SerialDataPresent */


static uint8_t GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycFr ] ;
  SerialService . RXDRec . CycFr ++ ;
  if ( SerialService . RXDRec . CycFr >= RXDBuffSize )
    SerialService . RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* GetSerial */


static void AddToCyclic ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  CyclicBuffer . Buffer [ CyclicBuffer . CycTo ] = Ch ;
  CyclicBuffer . CycTo ++ ;
//  if ( CyclicBuffer . CycTo >= CyclBuffSize )
//    CyclicBuffer. CycTo = 0 ;
} /* AddToCyclic */


static void AddToCyclicTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* AddToCyclicTxt */


static void AddToCyclicFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* AddToCyclicFlashTxt */


static uint8_t CyclicDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicBuffer . CycFr == CyclicBuffer . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* CyclicDataPresent */


static uint8_t GetFromCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = CyclicBuffer . Buffer [ CyclicBuffer . CycFr ] ;
  CyclicBuffer . CycFr ++ ;
//  if ( CyclicBuffer . CycFr >= RXDBuffSize )
//    CyclicBuffer . CycFr = 0 ;
  return ( Ch ) ;
} /* GetFromCyclic */


static void FlushCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  while ( CyclicDataPresent ( ) )
  {
    Ch = GetFromCyclic ( );
    SendSerial ( Ch ) ;
  } /* while */ ;
} /* FlushCyclic */


static void USARTHardwareInit ( uint16_t Speed )
{
  /*----------------------------------------------------------------*/
  UCSR0A = 0 ;
  UCSR0B = ( 1 << RXCIE0 ) | ( 1 << TXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) ;
  UCSR0C = ( 1 << UCSZ01 ) | ( 1 << UCSZ00 ) ;
  UBRR0 = Speed ;
} /* USARTHardwareInit */


static void HardwareInit ( uint16_t Speed )
{
  /*----------------------------------------------------------------*/
  TCCR0A = 0 ;
  TCCR0B = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TCNT0 = 0 ;
  OCR0A = 0 ;
  OCR0B = 0 ;
  TIMSK0 = ( 1 << TOIE0 ) ;
  DDRD &= ~ ( 1 << RxDPin ) ;
  DDRD |= ( 1 << TxDPin ) | ( 1 << RxTxDirPin ) ;
  USARTHardwareInit ( Speed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << StatusLed ) | ( 1 << RxTxLed ) | ( 1 << KeyLed ) ;
  LEDPortDirection |= Bits ;
//  LEDPort |= Bits ;
  LEDPort &= ~ Bits ;
  Bits = ( 1 << KeyPin ) ;
  KeyPortDirection &= ~ Bits ;
} /* EnvirInit */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static void Assign ( uint8_t * SubDeviceID )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* Assign */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
    {
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
} /* CompareFlash */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  WatchdogReset ;
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
} /* ReadEEBlock */


static void WriteEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  CodePassword ( ) ;
  WatchdogOff ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  WatchdogOn ;
  DecodePassword ( ) ;
} /* WriteEEBlock */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ Buffer [ Loop ] & 0x0F ] ;
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ ( Buffer [ Loop ] >> 4 ) & 0x0F ] ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void WordToStr ( uint8_t * Number ,
                        uint8_t   ConvWidth ,
                        uint8_t   EmptyCh ,
                        uint16_t  Value )
{
  uint8_t Loop ;
  uint8_t LocalBuffer [ 8 ] ;
  uint8_t LastDigit ;
  /*----------------------------------------------------------------*/
  if ( ConvWidth > 7 )
  {
    * Number ++ = '!' ;
    * Number = ( uint8_t ) 0 ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ConvWidth ; Loop ++ )
    LocalBuffer [ Loop ] = EmptyCh ;
  LocalBuffer [ ConvWidth ] = null ;
  Loop = ConvWidth - 1 ;
  for ( ; ; )
  {
    LastDigit = ( uint8_t ) ( Value % 10 ) ;
    LocalBuffer [ Loop ] = LastDigit + ( uint8_t ) 0x30 ;
    Value = Value / 10 ;
    if ( ( ! Loop ) || ( ! Value ) )
      break ;
    Loop -- ;
  } /* for */ ;
  for ( Loop = 0 ; Loop <= ConvWidth ; Loop ++ )
    * Number ++ = LocalBuffer [ Loop ] ;
} /* WordToStr */


static uint8_t * BegOfStr ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  while ( * String == Space )
    String ++ ;
  return ( String ) ;
} /* BegOfStr */


static uint8_t StringToNumb ( uint8_t * String )
{
  uint8_t Ch ;
  uint8_t RetValue ;
  /*----------------------------------------------------------------*/
  RetValue = 0 ;
  for ( ; ; )
  {
    Ch = * String ++ ;
    if ( ! Ch )
      break ;
    if ( ( Ch < '0' ) || ( Ch > '9' ) )
    {
      RetValue = 0 ;
      break ;
    } /* if */ ;
    RetValue = 10 * RetValue + Ch - 0x30 ;
  } /* for */ ;
  return ( RetValue ) ;
} /* StringToNumb */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static uint8_t MeDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SerialService . DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareFlash ( SerialService . DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void AddToCyclicReplyDevicesID ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( SerialService . SourceDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( SerialService . DestinationDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
} /* AddToCyclicReplyDevicesID */


static void AddToCyclicReplyForCommand ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclic ( 'R' ) ;
  AddToCyclicTxt ( SerialService . CommandText ) ;
  AddToCyclic ( ParamSeparator ) ;
} /* AddToCyclicReplyForCommand */


static void AddToCyclicCommandReplyOK ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) OK ) ;
} /* AddToCyclicCommandReplyOK */


static void AddToCyclicCommandReplyFail ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) Fail ) ;
} /* AddToCyclicCommandReplyFail */


static void ConnectCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( SerialService . ParamText [ Loop ] != Ch )
      return ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ConnectCommandExecute */


static void DisConnectCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  AddToCyclicCommandReplyOK ( ) ;
} /* DisConnectCommandExecute */


static void SetDeviceNameCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . MyDeviceID ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SetDeviceNameCommandExecute */


static void HelloCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) DeviceName ) ;
} /* HelloCommandExecute */


static void GetVersionCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) HardwareSoftwareVersion ) ;
} /* GetVersionCommandExecute */


static void ReinitCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlock . CRCValue = 0 ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  WriteEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SaveEECommandExecute */


static void AddSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( ! EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      Assign ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddSuperiorChnCommandExecute */


static void DelSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . EventDestinDeviceID [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( EEBlock . EventDestinDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelSuperiorChnCommandExecute */


static void ListSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ListSuperiorChnCommandExecute */


static void AddStatusChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < StatusDestinDeviceArraySize ; Loop ++ )
  {
    if ( ! EEBlock . StatusDestinDeviceID [ Loop ] [ 0 ] )
    {
      Assign ( EEBlock . StatusDestinDeviceID [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddStatusChnCommandExecute */


static void DelStatusChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < StatusDestinDeviceArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . StatusDestinDeviceID [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( EEBlock . StatusDestinDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelStatusChnCommandExecute */


static void ListStatusChnCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < StatusDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . StatusDestinDeviceID [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < StatusDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . StatusDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( EEBlock . StatusDestinDeviceID [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ListStatusChnCommandExecute */


static void ChStateServiceOnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( ! TempEchoDevices [ Loop ] [ 0 ] )
    {
      Assign ( TempEchoDevices [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOnCommandExecute */


static void ChStateServiceOffCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( TempEchoDevices [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOffCommandExecute */


static void ChStateServiceListCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( TempEchoDevices [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ChStateServiceListCommandExecute */


static uint8_t CommandRequiresParameters ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  if ( CompareFlash ( String , ( uint16_t ) ChOnCommand ) )
    return ( false ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChOffCommand ) )
    return ( false ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChInvCommand ) )
    return ( false ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOffCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOffCommand ) )
    return ( true ) ;
  return ( false ) ;
} /* CommandRequiresParameters */


static uint8_t DecodeCommandName ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  if ( CompareFlash ( String , ( uint16_t ) ChOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChOffCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChInvCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOffCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOffCommand ) )
    return ( true ) ;
  return ( false ) ;
} /* DecodeCommandName */


static uint8_t DecodeOperation ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  if ( CompareFlash ( String , ( uint16_t ) StartSelector ) )
    return ( StartActiveKeyCode ) ;
  if ( CompareFlash ( String , ( uint16_t ) StopSelector ) )
    return ( StopActiveKeyCode ) ;
  return ( 0xFF ) ;
} /* DecodeOperation */


static uint8_t UnpackData ( KeyCommParamRecT * KeyCmmDescr )
{
  CommandBufferType Tmp ;
  uint8_t Loop ;
  uint8_t Ch ;
  uint8_t ArrIndex ;
  /*----------------------------------------------------------------*/
  ClearBlock ( KeyCmmDescr , sizeof ( KeyCommParamRecT ) ) ;
  ClearBlock ( Tmp , sizeof ( CommandBufferType ) ) ;
  ArrIndex = 0 ;
  for ( Loop = 0 ; Loop < CommandBufferSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ ArrIndex ++ ] ;
    if ( ( Ch < 0x20 ) || ( Ch == ParamSeparator ) )
    {
      break ;
    } /* if */ ;
    Tmp [ Loop ] = Ch ;
  } /* for */ ;
  KeyCmmDescr -> CommKey = DecodeOperation ( Tmp ) ;
  if ( KeyCmmDescr -> CommKey == 0xFF )
  {
    return ( false ) ;
  } /* if */ ;
  if ( Ch != ParamSeparator )
  {
    return ( false ) ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < CommandBufferSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ ArrIndex ++ ] ;
    if ( ( Ch < 0x20 ) || ( Ch == ParamSeparator ) )
    {
      break ;
    } /* if */ ;
    KeyCmmDescr -> CommDescr . CommandText [ Loop ] = Ch ;
  } /* for */ ;
  if ( ! DecodeCommandName ( KeyCmmDescr -> CommDescr . CommandText ) )
  {
    return ( false ) ;
  } /* if */ ;
  if ( Ch == ParamSeparator )
  {
    if ( ! CommandRequiresParameters ( KeyCmmDescr -> CommDescr . CommandText ) )
    {
      return ( false ) ;
    } /* if */ ;
    ClearBlock ( Tmp , sizeof ( CommandBufferType ) ) ;
    for ( Loop = 0 ; Loop < CommandBufferSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ ArrIndex ++ ] ;
      if ( Ch < 0x20 )
      {
        break ;
      } /* if */ ;
      Tmp [ Loop ] = Ch ;
    } /* for */ ;
    KeyCmmDescr -> CommDescr . ParamValue = StringToNumb ( Tmp ) ;
    if ( ! KeyCmmDescr -> CommDescr . ParamValue )
    {
      return ( false ) ;
    } /* if */ ;
  } /* if ... */
  else
  {
    if ( CommandRequiresParameters ( KeyCmmDescr -> CommDescr . CommandText ) )
    {
      return ( false ) ;
    } /* if */ ;
  } /* if ... else */ ;
  return ( true ) ;
} /* UnpackData */


static void AddKeyCommandExecute ( void )
{
  KeyCommParamRecT AddElement ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! UnpackData ( & AddElement ) )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  if ( AddElement . CommKey == StartActiveKeyCode )
  {
    EEBlock . KeyStartService = AddElement . CommDescr ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  if ( AddElement . CommKey == StopActiveKeyCode )
  {
    EEBlock . KeyStopService = AddElement . CommDescr ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddKeyCommandExecute */


static void DelKeyCommandExecute ( void )
{
  KeyCommParamRecT AddElement ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! UnpackData ( & AddElement ) )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  if ( AddElement . CommKey == StartActiveKeyCode )
  {
    ClearBlock ( & EEBlock . KeyStartService , sizeof ( KeyCommDescrRecT ) ) ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  if ( AddElement . CommKey == StopActiveKeyCode )
  {
    ClearBlock ( & EEBlock . KeyStopService , sizeof ( KeyCommDescrRecT ) ) ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelKeyCommandExecute */


static void DisplayKeyElement ( PtrKeyCommDescrRecT CommDescr ,
                                uint16_t            Selector )
{
  uint8_t Number [ 8 ] ;
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( Selector ) ;
  AddToCyclic ( ParamSeparator ) ;
  if ( CommDescr -> CommandText [ 0 ] )
  {
    AddToCyclicTxt ( CommDescr -> CommandText ) ;
    if ( CommDescr -> ParamValue )
    {
      AddToCyclic ( ParamSeparator ) ;
      WordToStr ( Number , 3 , Space , CommDescr -> ParamValue ) ;
      AddToCyclicTxt ( BegOfStr ( Number ) ) ;
    } /* if */ ;
    AddToCyclic ( cr ) ;
  } /* if ... */
  else
  {
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
  } /* if ... else */ ;
} /* DisplayKeyElement */


static void ListKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  DisplayKeyElement ( & EEBlock . KeyStartService , ( uint16_t ) StartSelector ) ;
  DisplayKeyElement ( & EEBlock . KeyStopService , ( uint16_t ) StopSelector ) ;
} /* ListKeyCommandExecute */


static void NewPasswordCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( SerialService . ParamText [ 0 ] )
  {
    for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ Loop ] ;
      EEBlock . Password [ Loop ] = Ch ;
      if ( Ch == null )
        break ;
    } /* for */ ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* NewPasswordCommandExecute */


static void SingleSendKeyEvent ( uint8_t * DestinationDeviceID ,
                                 uint8_t * DeviceCommandText ,
                                 uint8_t   DeviceParamValue )
{
  uint8_t Number [ 6 ] ;
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( DestinationDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( DeviceCommandText ) ;
  if ( DeviceParamValue )
  {
    WordToStr ( Number , 3 , ' ' , DeviceParamValue ) ;
    AddToCyclic ( ParamSeparator ) ;
    AddToCyclicTxt ( BegOfStr ( Number ) ) ;
  } /* if */ ;
  AddToCyclic ( cr ) ;
} /* SingleSendKeyEvent */


static void SendKeyEvents ( uint8_t * CommandText ,
                            uint8_t   ParamValue )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( CommandText [ 0 ] )
  {
    for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
    {
      if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
      {
        SingleSendKeyEvent ( EEBlock . EventDestinDeviceID [ Loop ] , CommandText , ParamValue ) ;
      } /* if */ ;
    } /* for */ ;
    for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
    {
      if ( TempEchoDevices [ Loop ] [ 0 ] )
      {
        SingleSendKeyEvent ( TempEchoDevices [ Loop ] , CommandText , ParamValue ) ;
      } /* if */ ;
    } /* for */ ;
  } /* if */ ;
} /* SendKeyEvents */


static void SendStatusEvents ( uint8_t Status )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < StatusDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . StatusDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddToCyclicTxt ( EEBlock . StatusDestinDeviceID [ Loop ] ) ;
      AddToCyclic ( PartSeparator ) ;
      AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
      AddToCyclic ( PartSeparator ) ;
      if ( Status )
        AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOnState ) ;
      else
        AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOffState ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* SendStatusEvents */


static void ReplyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicDataPresent ( ) )
  {
    FlushCyclic ( ) ;
  } /* if ... */
  else
  {
    SendSerialTxt ( SerialService . SourceDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialTxt ( SerialService . DestinationDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialFlashTxt ( ( uint16_t ) NoDataText ) ;
  } /* if ... else */ ;
} /* ReplyCommandExecute */


static void ExecuteAssociatonCommand ( uint8_t *      String ,
                                       uint8_t        AssociationTableSize ,
                                       uint16_t       AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* ExecuteAssociatonCommand */


static void ExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( ) ;
    if ( FrameToMe )
      StrCopy ( SerialService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    ExecuteAssociatonCommand ( SerialService . CommandText ,
                               CommandNameAssociationTabeSize ,
                               ( uint16_t ) CommandNameAssociationTabe ) ;
  } /* if */ ;
} /* ExecuteCommand */


static void ResetOperation ( void ) ;

static void DestinationIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . DestDeviceNameIndex < DeviceIDSize )
  {
    SerialService . DestinationDeviceName [ SerialService . DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . SrcDeviceNameIndex < DeviceIDSize )
  {
    SerialService . SourceDeviceName [ SerialService . SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . CommandTextIndex < CommandBufferSize )
  {
    SerialService . CommandText [ SerialService . CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . ParamTextIndex < ParamBufferSize )
  {
    SerialService . ParamText [ SerialService . ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( null ) ;
      SerialService . SerialService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( null ) ;
      SerialService . SerialService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( null ) ;
      SerialService . SerialService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static void ExecuteKeyCommand ( uint8_t KeyC )
{
  /*----------------------------------------------------------------*/
  switch ( KeyC )
  {
    case StartActiveKeyCode      :
      SendKeyEvents ( EEBlock . KeyStartService . CommandText ,
                      EEBlock . KeyStartService . ParamValue ) ;
      SendStatusEvents ( true ) ;
      break ;
    case StopActiveKeyCode       :
      SendKeyEvents ( EEBlock . KeyStopService . CommandText ,
                      EEBlock . KeyStopService . ParamValue ) ;
      SendStatusEvents ( false ) ;
      break ;
    default                 :
      return ;
  } /* switch */ ;
} /* ExecuteKeyCommand */


static void ResetOperation ( void )
{
  /*----------------------------------------------------------------*/
  ClearBlock ( SerialService . CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( SerialService . ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( SerialService . DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( SerialService . SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  SerialService . CommandTextIndex = 0 ;
  SerialService . ParamTextIndex = 0 ;
  SerialService . DestDeviceNameIndex = 0 ;
  SerialService . SrcDeviceNameIndex = 0 ;
  SerialService . SerialService = SelectDestOperation ;
} /* ResetOperation */


static void SoftwareInit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  TimerCounter = 0 ;
  CyclicBuffer . CycFr = 0 ;
  CyclicBuffer . CycTo = 0 ;
  SerialService . RXDRec . CycFr = 0 ;
  SerialService . RXDRec . CycTo = 0 ;
  SerialService . TXDRec . CycFr = 0 ;
  SerialService . TXDRec . CycTo = 0 ;
  SerialService . TXDRec . State = IdleState ;
  SerialService . SerialLEDDelay = 0 ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
  } /* for */ ;
  KeyStatus = KeybIdle ;
  KeybCodeBuffer = 0 ;
  KeyDelayCt = 0 ;
  ResetOperation ( ) ;
  KeybDelayTimer = 0 ;
  TimeEvent = false ;
} /* SoftwareInit */


static void LEDTest ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << StatusLed ) | ( 1 << RxTxLed ) | ( 1 << KeyLed ) ;
  LEDPort &= ~ Bits ;
  TimeEvent = 0 ;
  TimerCounter = 0 ;
  for ( ; ; )
  {
     WatchdogReset ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  LEDPort |= Bits ;
} /* LEDTest */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( Speed38400bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  USARTReceiveMode ( ) ;
  EEBlockOperation ( ) ;
  WatchdogOn ;
  sei ( ) ;
  LEDTest ( ) ;
  for ( ; ; )
  {
    WatchdogReset ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      LEDPort ^= 1 << StatusLed ;
    } /* if */ ;
    if ( SerialService . SerialService )
      SerialService . SerialService ( ) ;
    if ( KeyPressed ( ) )
    {
      ExecuteKeyCommand ( ReadKey ( ) ) ;
    } /* if */ ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


